/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#include <ntddk.h>
#include "mx_auto_config.h"
#include "mx_int.h"
#include "mcp_types.h"
#include "mx_ntddk.h"
#include "mx_arch_io.h"
#undef SLIST_ENTRY
#include "bsd/queue.h"

struct mx__pa_hash_entry
{
  SLIST_ENTRY(mx__pa_hash_entry) entries;
  int ref_count;
  PMDL mdl;
  PHYSICAL_ADDRESS pa;
};

SLIST_HEAD(mx__pa_hash_head, mx__pa_hash_entry);

/* Shrink the hash table to artificially force collisions. */
#define PA_HASH_TABLE_SHIFT 16
#define PA_HASH_TABLE_SIZE (1<<PA_HASH_TABLE_SHIFT)
FAST_MUTEX g_pa_mutex;
struct mx__pa_hash_head *g_pa_hash;
NPAGED_LOOKASIDE_LIST g_pa_hash_lookaside;

/* For some reason, calling KeInitializeSemaphore in mx.c results in a
   link error. It may be a header issue (no function prototype?) so we
   wrap it here. Oddly enough, KeReleaseSemaphore is ok. */
void mx_init_sem(KSEMAPHORE* sem)
{
  KeInitializeSemaphore(sem, 0, 0x7fffffff);
}

/*
-------------------------------------------------------------------------------
lookup3.c, by Bob Jenkins, May 2006, Public Domain.

These are functions for producing 32-bit hashes for hash table lookup.
hashword(), hashlittle(), hashlittle2(), hashbig(), mix(), and final() 
are externally useful functions.  Routines to test the hash are included 
if SELF_TEST is defined.  You can use this free for any purpose.  It's in
the public domain.  It has no warranty.

You probably want to use hashlittle().  hashlittle() and hashbig()
hash byte arrays.  hashlittle() is is faster than hashbig() on
little-endian machines.  Intel and AMD are little-endian machines.
On second thought, you probably want hashlittle2(), which is identical to
hashlittle() except it returns two 32-bit hashes for the price of one.  
You could implement hashbig2() if you wanted but I haven't bothered here.

If you want to find a hash of, say, exactly 7 integers, do
  a = i1;  b = i2;  c = i3;
  mix(a,b,c);
  a += i4; b += i5; c += i6;
  mix(a,b,c);
  a += i7;
  final(a,b,c);
then use c as the hash value.  If you have a variable length array of
4-byte integers to hash, use hashword().  If you have a byte array (like
a character string), use hashlittle().  If you have several byte arrays, or
a mix of things, see the comments above hashlittle().  

Why is this so big?  I read 12 bytes at a time into 3 4-byte integers, 
then mix those integers.  This is fast (you can do a lot more thorough
mixing with 12*3 instructions on 3 integers than you can with 3 instructions
on 1 byte), but shoehorning those bytes into integers efficiently is messy.
-------------------------------------------------------------------------------
*/

/*
 * My best guess at if you are big-endian or little-endian.  This may
 * need adjustment.
 */
#define HASH_LITTLE_ENDIAN 1
#define HASH_BIG_ENDIAN 0

#define hashsize(n) ((uint32_t)1<<(n))
#define hashmask(n) (hashsize(n)-1)
#define rot(x,k) (((x)<<(k)) | ((x)>>(32-(k))))

/*
-------------------------------------------------------------------------------
mix -- mix 3 32-bit values reversibly.

This is reversible, so any information in (a,b,c) before mix() is
still in (a,b,c) after mix().

If four pairs of (a,b,c) inputs are run through mix(), or through
mix() in reverse, there are at least 32 bits of the output that
are sometimes the same for one pair and different for another pair.
This was tested for:
* pairs that differed by one bit, by two bits, in any combination
  of top bits of (a,b,c), or in any combination of bottom bits of
  (a,b,c).
* "differ" is defined as +, -, ^, or ~^.  For + and -, I transformed
  the output delta to a Gray code (a^(a>>1)) so a string of 1's (as
  is commonly produced by subtraction) look like a single 1-bit
  difference.
* the base values were pseudorandom, all zero but one bit set, or 
  all zero plus a counter that starts at zero.

Some k values for my "a-=c; a^=rot(c,k); c+=b;" arrangement that
satisfy this are
    4  6  8 16 19  4
    9 15  3 18 27 15
   14  9  3  7 17  3
Well, "9 15 3 18 27 15" didn't quite get 32 bits diffing
for "differ" defined as + with a one-bit base and a two-bit delta.  I
used http://burtleburtle.net/bob/hash/avalanche.html to choose 
the operations, constants, and arrangements of the variables.

This does not achieve avalanche.  There are input bits of (a,b,c)
that fail to affect some output bits of (a,b,c), especially of a.  The
most thoroughly mixed value is c, but it doesn't really even achieve
avalanche in c.

This allows some parallelism.  Read-after-writes are good at doubling
the number of bits affected, so the goal of mixing pulls in the opposite
direction as the goal of parallelism.  I did what I could.  Rotates
seem to cost as much as shifts on every machine I could lay my hands
on, and rotates are much kinder to the top and bottom bits, so I used
rotates.
-------------------------------------------------------------------------------
*/
#define mix(a,b,c) \
{ \
  a -= c;  a ^= rot(c, 4);  c += b; \
  b -= a;  b ^= rot(a, 6);  a += c; \
  c -= b;  c ^= rot(b, 8);  b += a; \
  a -= c;  a ^= rot(c,16);  c += b; \
  b -= a;  b ^= rot(a,19);  a += c; \
  c -= b;  c ^= rot(b, 4);  b += a; \
}

/*
-------------------------------------------------------------------------------
final -- final mixing of 3 32-bit values (a,b,c) into c

Pairs of (a,b,c) values differing in only a few bits will usually
produce values of c that look totally different.  This was tested for
* pairs that differed by one bit, by two bits, in any combination
  of top bits of (a,b,c), or in any combination of bottom bits of
  (a,b,c).
* "differ" is defined as +, -, ^, or ~^.  For + and -, I transformed
  the output delta to a Gray code (a^(a>>1)) so a string of 1's (as
  is commonly produced by subtraction) look like a single 1-bit
  difference.
* the base values were pseudorandom, all zero but one bit set, or 
  all zero plus a counter that starts at zero.

These constants passed:
 14 11 25 16 4 14 24
 12 14 25 16 4 14 24
and these came close:
  4  8 15 26 3 22 24
 10  8 15 26 3 22 24
 11  8 15 26 3 22 24
-------------------------------------------------------------------------------
*/
#define final(a,b,c) \
{ \
  c ^= b; c -= rot(b,14); \
  a ^= c; a -= rot(c,11); \
  b ^= a; b -= rot(a,25); \
  c ^= b; c -= rot(b,16); \
  a ^= c; a -= rot(c,4);  \
  b ^= a; b -= rot(a,14); \
  c ^= b; c -= rot(b,24); \
}

/*
--------------------------------------------------------------------
 This works on all machines.  To be useful, it requires
 -- that the key be an array of uint32_t's, and
 -- that the length be the number of uint32_t's in the key

 The function hashword() is identical to hashlittle() on little-endian
 machines, and identical to hashbig() on big-endian machines,
 except that the length has to be measured in uint32_ts rather than in
 bytes.  hashlittle() is more complicated than hashword() only because
 hashlittle() has to dance around fitting the key bytes into registers.
--------------------------------------------------------------------
*/
uint32_t hashword(
const uint32_t *k,                   /* the key, an array of uint32_t values */
size_t          length,               /* the length of the key, in uint32_ts */
uint32_t        initval)         /* the previous hash, or an arbitrary value */
{
  uint32_t a,b,c;

  /* Set up the internal state */
  a = b = c = 0xdeadbeef + (((uint32_t)length)<<2) + initval;

  /*------------------------------------------------- handle most of the key */
  while (length > 3)
  {
    a += k[0];
    b += k[1];
    c += k[2];
    mix(a,b,c);
    length -= 3;
    k += 3;
  }

  /*------------------------------------------- handle the last 3 uint32_t's */
  switch(length)                     /* all the case statements fall through */
  { 
  case 3 : c+=k[2];
  case 2 : b+=k[1];
  case 1 : a+=k[0];
    final(a,b,c);
  case 0:     /* case 0: nothing left to add */
    break;
  }
  /*------------------------------------------------------ report the result */
  return c;
}

int
mx__pa_hash_init(void)
{
  int i;

  ExInitializeFastMutex(&g_pa_mutex);
  g_pa_hash = ExAllocatePoolWithTag(NonPagedPool,
				    PA_HASH_TABLE_SIZE * sizeof (*g_pa_hash),
				    'HSAH');
  if (g_pa_hash == NULL) {
    goto abort_with_nothing;
  }
  for (i = 0; i < PA_HASH_TABLE_SIZE; ++i) {
    SLIST_INIT(&g_pa_hash[i]);
  }
  ExInitializeNPagedLookasideList(&g_pa_hash_lookaside, NULL, NULL, 0,
				  sizeof (struct mx__pa_hash_entry),
				  'HSAH', 0);
  return 0;

 abort_with_nothing:
  return 1;
}

void
mx__pa_hash_fini(void)
{
  ExDeleteNPagedLookasideList(&g_pa_hash_lookaside);
  ExFreePoolWithTag(g_pa_hash, 'HSAH');
}

/* Looks up key in hash table. If not found, and create is 1, allocate
 * a new entry and set the ref count to 0. */
struct mx__pa_hash_entry *
mx__pa_hash_lookup(struct mx__pa_hash_head *hash_table,
		   uint32_t table_size, uint64_t key, int create)
{
  uint32_t idx;
  struct mx__pa_hash_entry *e;
  int found = 0;

  idx = hashword(&key, sizeof (key)/sizeof (uint32_t), 0) & hashmask(PA_HASH_TABLE_SHIFT);
  SLIST_FOREACH(e, &hash_table[idx], entries) {
    if ((uintptr_t)e->pa.QuadPart == key) {
      found = 1;
      break;
    }
  }
  if (!found && create) {
    e = ExAllocateFromNPagedLookasideList(&g_pa_hash_lookaside);
    if (e) {
      e->ref_count = 0;
      SLIST_INSERT_HEAD(&hash_table[idx], e, entries);
    }
  }

  return e;
}

void
mx__pa_hash_remove(struct mx__pa_hash_head *hash_table,
		   uint32_t table_size, uint64_t key)
{
  uint32_t idx;
  struct mx__pa_hash_entry *p, *e;
  int found = 0;

  idx = hashword(&key, sizeof (key)/sizeof (uint32_t), 0) & hashmask(PA_HASH_TABLE_SHIFT);
  p = NULL;
  SLIST_FOREACH(e, &hash_table[idx], entries) {
    if ((uintptr_t)e->pa.QuadPart == key) {
      found = 1;
      break;
    }
    p = e;
  }
  if (found) {
    if (p == NULL) {
      /* First entry. */
      SLIST_REMOVE_HEAD(&hash_table[idx], entries);
    }
    else {
      SLIST_NEXT(p, entries) = SLIST_NEXT(SLIST_NEXT(p, entries), entries);
    }
    ExFreeToNPagedLookasideList(&g_pa_hash_lookaside, e);
  }
  else {
  }
}

int 
mx_pin_page(mx_instance_state_t *is, mx_page_pin_t *pin, int flags, uint64_t memory_context)
{
  PMDL ro_mdl;
  PHYSICAL_ADDRESS pa;
  struct mx__pa_hash_entry *he;
  PMDL rw_mdl;

  ExAcquireFastMutex(&g_pa_mutex);

  ro_mdl = IoAllocateMdl((void*)(uintptr_t)pin->va, PAGE_SIZE, FALSE,
			 FALSE, NULL);
  if (ro_mdl == NULL) {
    goto abort_with_nothing;
  }
  __try {
    MmProbeAndLockPages(ro_mdl, KernelMode, IoReadAccess);
  }
  __except (EXCEPTION_EXECUTE_HANDLER) {
    goto abort_with_ro_mdl;
  }
  pa = MmGetPhysicalAddress((void*)(uintptr_t)pin->va);
  he = mx__pa_hash_lookup(g_pa_hash, PA_HASH_TABLE_SIZE, pa.QuadPart, 1);
  /* he is created and ref_count initialized to 0 if not found. */
  if (he == NULL) {
    goto abort_with_pa;
  }
  else if (he->ref_count == 0) {
    rw_mdl = IoAllocateMdl((void*)(uintptr_t)pin->va, PAGE_SIZE, FALSE,
			   FALSE, NULL);
    if (rw_mdl == NULL) {
      goto abort_with_he;
    }
    __try {
      MmProbeAndLockPages(rw_mdl, KernelMode, IoModifyAccess);
    }
    __except (EXCEPTION_EXECUTE_HANDLER) {
      goto abort_with_rw_mdl;
    }

    /* Can't fail now. */
    he->ref_count++;
    he->mdl = rw_mdl;
    he->pa.QuadPart = pa.QuadPart;
  }
  else {
    he->ref_count++;
  }
  pin->dma.low = pa.LowPart;
  pin->dma.high = pa.HighPart;
  pin->mdl = he->mdl;

  MmUnlockPages(ro_mdl);
  IoFreeMdl(ro_mdl);

  ExReleaseFastMutex(&g_pa_mutex);

  return 0;

 abort_with_rw_mdl:
  IoFreeMdl(rw_mdl);
 abort_with_he:
  mx__pa_hash_remove(g_pa_hash, PA_HASH_TABLE_SIZE, pa.QuadPart);
 abort_with_pa:
  MmUnlockPages(ro_mdl);
 abort_with_ro_mdl:
  IoFreeMdl(ro_mdl);
 abort_with_nothing:
  ExReleaseFastMutex(&g_pa_mutex);
  return EFAULT;
}

void
mx_unpin_page(mx_instance_state_t *is, mx_page_pin_t *pin, int flags)
{
  struct mx__pa_hash_entry *he;
  uint64_t pa64;

  ExAcquireFastMutex(&g_pa_mutex);

  pa64 = ((((uint64_t)pin->dma.high) << 32) |
	  pin->dma.low);
  he = mx__pa_hash_lookup(g_pa_hash, PA_HASH_TABLE_SIZE, pa64, 0);
  /* Don't create if not found. */
  if (he == NULL) {
  }
  else if (he->ref_count == 1) {
    /* This is the last unpin. Time to really do work. */
    mx__pa_hash_remove(g_pa_hash, PA_HASH_TABLE_SIZE, pa64);
    MmUnlockPages(pin->mdl);
    IoFreeMdl(pin->mdl);
  }
  else {
    he->ref_count--;
  }

  ExReleaseFastMutex(&g_pa_mutex);
}

#define MX_SCAN_PRIVILEGES 0
#if MX_SCAN_PRIVILEGES
#define MXSTR(x_) #x_
#define MX_CHECK(irp_, x_) \
do { \
  LUID luid; \
  BOOLEAN b; \
  luid = RtlConvertLongToLuid(x_); \
  b = SeSinglePrivilegeCheck(luid, irp_->RequestorMode); \
  MX_INFO((MXSTR(x_) " = %d\n", (int) b)); \
} while(0)
#endif

int mx_can_be_privileged(PIRP irp)
{
  LUID lock_memory_luid;
  LUID load_driver_luid;
  LUID impersonate_luid;
  BOOLEAN b;

#if MX_SCAN_PRIVILEGES
  MX_CHECK(irp, SE_MIN_WELL_KNOWN_PRIVILEGE);
  MX_CHECK(irp, SE_CREATE_TOKEN_PRIVILEGE);
  MX_CHECK(irp, SE_ASSIGNPRIMARYTOKEN_PRIVILEGE);
  MX_CHECK(irp, SE_LOCK_MEMORY_PRIVILEGE);
  MX_CHECK(irp, SE_INCREASE_QUOTA_PRIVILEGE);

  MX_CHECK(irp, SE_MACHINE_ACCOUNT_PRIVILEGE);
  MX_CHECK(irp, SE_TCB_PRIVILEGE);
  MX_CHECK(irp, SE_SECURITY_PRIVILEGE);
  MX_CHECK(irp, SE_TAKE_OWNERSHIP_PRIVILEGE);
  MX_CHECK(irp, SE_LOAD_DRIVER_PRIVILEGE);
  MX_CHECK(irp, SE_SYSTEM_PROFILE_PRIVILEGE);
  MX_CHECK(irp, SE_SYSTEMTIME_PRIVILEGE);
  MX_CHECK(irp, SE_PROF_SINGLE_PROCESS_PRIVILEGE);
  MX_CHECK(irp, SE_INC_BASE_PRIORITY_PRIVILEGE);
  MX_CHECK(irp, SE_CREATE_PAGEFILE_PRIVILEGE);
  MX_CHECK(irp, SE_CREATE_PERMANENT_PRIVILEGE);
  MX_CHECK(irp, SE_BACKUP_PRIVILEGE);
  MX_CHECK(irp, SE_RESTORE_PRIVILEGE);
  MX_CHECK(irp, SE_SHUTDOWN_PRIVILEGE);
  MX_CHECK(irp, SE_DEBUG_PRIVILEGE);
  MX_CHECK(irp, SE_AUDIT_PRIVILEGE);
  MX_CHECK(irp, SE_SYSTEM_ENVIRONMENT_PRIVILEGE);
  MX_CHECK(irp, SE_CHANGE_NOTIFY_PRIVILEGE);
  MX_CHECK(irp, SE_REMOTE_SHUTDOWN_PRIVILEGE);
  MX_CHECK(irp, SE_UNDOCK_PRIVILEGE);
  MX_CHECK(irp, SE_SYNC_AGENT_PRIVILEGE);
  MX_CHECK(irp, SE_ENABLE_DELEGATION_PRIVILEGE);
  MX_CHECK(irp, SE_MANAGE_VOLUME_PRIVILEGE);
  MX_CHECK(irp, SE_IMPERSONATE_PRIVILEGE);
  /*MX_CHECK(irp, SE_CREATE_GLOBAL_PRIVILEGE);*/
#endif

  lock_memory_luid = RtlConvertLongToLuid(SE_LOCK_MEMORY_PRIVILEGE);
  load_driver_luid = RtlConvertLongToLuid(SE_LOAD_DRIVER_PRIVILEGE);
  impersonate_luid = RtlConvertLongToLuid(SE_IMPERSONATE_PRIVILEGE);
  b = (SeSinglePrivilegeCheck(lock_memory_luid, irp->RequestorMode) ||
       SeSinglePrivilegeCheck(load_driver_luid, irp->RequestorMode) ||
       SeSinglePrivilegeCheck(impersonate_luid, irp->RequestorMode));
  return (int)b;
}

ULONG
mx_wrap_hal_get(IN ULONG BusNumber,
		IN ULONG SlotNumber,
		IN PVOID Buffer,
		IN ULONG Offset,
		IN ULONG Length)
{
  return HalGetBusDataByOffset(PCIConfiguration,
			       BusNumber,
			       SlotNumber,
			       Buffer,
			       Offset,
			       Length);
}

ULONG
mx_wrap_hal_set(IN ULONG  BusNumber,
		IN ULONG  SlotNumber,
		IN PVOID  Buffer,
		IN ULONG  Offset,
		IN ULONG  Length)
{
  return HalSetBusDataByOffset(PCIConfiguration,
			       BusNumber,
			       SlotNumber,
			       Buffer,
			       Offset,
			       Length);
}
