#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>

#include "mx_auto_config.h"
#include "mx__shim.h"
#include "myriexpress.h"
#include "mx_byteswap.h"
#include "mx_raw.h"
#include "mx__thread.h"
#if MX_OS_WINNT
#define random rand
#define srandom srand
#endif

#define MXOED_DEBUG 0

#define MAX_PEERS 8192
#define MAX_NICS 8
#define MXOE_PORT 2314
#define MAX_IFC_CNT 16
#define BROADCAST_INTERVAL 1000
#define LONG_BROADCAST_INTERVAL 180000
#define BROADCAST_COUNT 8

#define ETHER_TYPE_MX	0x86DF
#define MYRI_TYPE_ETHER 0x0009

/*
 * Info about each NIC
 */
struct mxoed_pkt {
  uint32_t dest_mac_high32;
  uint16_t dest_mac_low16;
  uint16_t src_mac_high16;
  uint32_t src_mac_low32;
  uint16_t proto;		/* ethertype */
  uint16_t sender_peer_index;
  uint8_t pkt_type;
  uint8_t chargap[3];
  uint32_t gap[3];		/* pad to 32 bytes */
  uint32_t nic_id_hi;
  uint32_t nic_id_lo;
  uint32_t serial;
  uint8_t  pad[20];		/* then to 64 bytes */
};

struct mxoed_eom_pkt {
  uint16_t pad;
  uint16_t myri_type;
  uint16_t subtype;
  uint16_t payload_len;
  uint16_t seqnum;
  uint16_t gw_high16;
  uint32_t gw_low32;
  struct mxoed_pkt mxoe;
};

struct nic_info {
  mx_raw_endpoint_t raw_ep;		/* endpoint for this NIC */

  int nic_index;
  uint32_t net_type;
  uint64_t my_nic_id;
  uint32_t my_serial;

  uint64_t peers[MAX_PEERS];
  uint64_t gw[MAX_PEERS];
  uint32_t peer_serial[MAX_PEERS];
  int num_peers;

  int bc_count;
  int bc_interval;

  struct mxoed_pkt outpkt;
  union {
    struct mxoed_pkt mxoe;
    struct mxoed_eom_pkt eom;
  } pkt;

  int die; /* set to non-zero to exit */
};


/* return difference between two timevals in ms (t1 - t2) */
static inline int32_t
tv_diff(
  struct timeval *t1,
  struct timeval *t2)
{
  int32_t ms;

  ms = (t1->tv_sec - t2->tv_sec) * 1000;
  ms += (t1->tv_usec - t2->tv_usec) / 1000;
  return ms;
}

/*
 * Define this peer to MX
 */
static void
write_peers_to_nic(
  struct nic_info *nip)
{
  mx_return_t rc;
  int i;

  rc = mx_raw_set_route_begin(nip->raw_ep);
  if (rc != MX_SUCCESS) {
    fprintf(stderr, "Error add peer init: %s\n", mx_strerror(rc));
    exit(1);
  }

  if (nip->net_type == MX_NET_ETHER) {
    /* write self */
    rc = mx_raw_set_route(nip->raw_ep, nip->my_nic_id,
			  NULL, 0, 0, 0, MX_HOST_MX);
    if (rc != MX_SUCCESS) {
      fprintf(stderr, "Error adding self to NIC %d: %s\n",
	      nip->nic_index, mx_strerror(rc));
      exit(1);
    }
  }
  
  /* add the peers */
  for (i = 0; i < nip->num_peers; i++) {
    if (nip->net_type == MX_NET_ETHER) {
      rc = mx_raw_set_route(nip->raw_ep, nip->peers[i],
			    NULL, 0, 0, 0, MX_HOST_MX);
    } else {
      uint8_t route[6];
      uint64_t gw;
      int j;

      gw = nip->gw[i];
      for (j=0; j < 6;j++)
	route[5 - j] = (gw >> (j * 8)) & 0xff;
      rc = mx_raw_set_route(nip->raw_ep, nip->peers[i],
			    route, 6, 0, 0, MX_HOST_MXvM);
    }
    if (rc != MX_SUCCESS) {
      fprintf(stderr, "Error adding peer to NIC %d: %s\n",
	  nip->nic_index, mx_strerror(rc));
      exit(1);
    }
  }

  rc = mx_raw_set_route_end(nip->raw_ep);
  if (rc != MX_SUCCESS) {
    fprintf(stderr, "Error add peer init: %s\n", mx_strerror(rc));
    exit(1);
  }

  rc = mx_raw_set_map_version(nip->raw_ep, 0, nip->my_nic_id, 1,
      nip->num_peers+1, 1);
  if (rc != MX_SUCCESS) {
    fprintf(stderr, "Error add peer init: %s\n", mx_strerror(rc));
    exit(1);
  }
}

void
add_peer(
  struct nic_info *nip,
  uint64_t peer_mac,
  int serial,
  uint64_t gw)
{

  /* Add this to our local peer table */
  nip->peers[nip->num_peers] = peer_mac;
  nip->gw[nip->num_peers] = gw;
  nip->peer_serial[nip->num_peers] = serial;
  ++nip->num_peers;

  write_peers_to_nic(nip);
}

int
get_peer_index(
  struct nic_info *nip,
  uint64_t peer_mac)
{
  int i;

  for (i=0; i<nip->num_peers; ++i) {
    if (nip->peers[i] == peer_mac) {
      return i;
    }
  }
  return -1;
}

void
broadcast_my_id(
  struct nic_info *nip)
{
  mx_return_t mxrc;

  mxrc = mx_raw_send(nip->raw_ep, 0, NULL, 0,
      &nip->outpkt, sizeof(nip->outpkt), NULL);
  if (mxrc != MX_SUCCESS) {
    fprintf(stderr, "Error sending raw packet: %s\n", mx_strerror(mxrc));
    exit(1);
  }
#if MXOED_DEBUG
    printf("sent my ID\n");
#endif
}

int
check_for_packet(
  struct nic_info *nip,
  int *elapsed)
{
  mx_return_t mxrc;
  mx_raw_status_t stat;
  void *context;
  uint32_t port;
  uint32_t len;
  struct timeval before;
  struct timeval after;
  int timeout;
  int rc;

  (void) mx_gettimeofday(&before, NULL);

  len = sizeof(nip->pkt);
  if (nip->bc_interval > 0) {
    timeout = nip->bc_interval;
  } else {
    timeout = 0;
  }
  mxrc = mx_raw_next_event(nip->raw_ep, &port, &context,
      &nip->pkt, &len, timeout, &stat);
  if (mxrc != MX_SUCCESS) {
    fprintf(stderr, "Error from mx_raw_next_event: %s\n", mx_strerror(mxrc));
    exit(1);
  }

  (void) mx_gettimeofday(&after, NULL);

  if (stat == MX_RAW_RECV_COMPLETE) {
#if MXOED_DEBUG
    int i;
    unsigned char *p = (unsigned char *)(&nip->pkt);

    printf("recv len = %d\n", len);
    for (i=0; i<16; ++i) printf(" %02x", p[i]); printf("\n");
    for (; i<32; ++i) printf(" %02x", p[i]); printf("\n");
    for (; i<48; ++i) printf(" %02x", p[i]); printf("\n");
#endif

    rc = 1;

  } else if (stat == MX_RAW_SEND_COMPLETE) {
#if MXOED_DEBUG
  printf("send complete\n");
#endif
    rc = 0;

  } else {
    rc = 0;
  }

  /* return elapsed time in ms */
  *elapsed = tv_diff(&after, &before);

#if MXOED_DEBUG
  printf("elapsed = %d\n", *elapsed);
#endif

  return rc;
}

void
process_pkt(
  struct nic_info *nip)
{
  uint32_t nic_half;
  uint64_t nic_id;
  int serial;
  int index;
  struct mxoed_pkt *pkt;
  uint64_t gw;

  if (nip->net_type == MX_NET_MYRI) {
    memmove((char*)&nip->pkt + 2, &nip->pkt, sizeof(nip->pkt) - 2);
    if (ntohs(nip->pkt.eom.myri_type) !=  MYRI_TYPE_ETHER
	|| ntohs(nip->pkt.eom.payload_len) !=  sizeof(nip->pkt.mxoe)
	|| ntohs(nip->pkt.eom.mxoe.proto) !=  ETHER_TYPE_MX)
      return;
    
    pkt = &nip->pkt.eom.mxoe;
    gw = (((uint64_t)ntohs(nip->pkt.eom.gw_high16) << 32)
	  + ntohl(nip->pkt.eom.gw_low32));
  } else {
    pkt = &nip->pkt.mxoe;
    gw = 0;
  }
  /* get peer NIC id from packet */
  nic_id = ntohl(pkt->nic_id_lo);
  nic_half = ntohl(pkt->nic_id_hi);
  nic_id |= ((uint64_t)nic_half << 32);

  serial = ntohl(pkt->serial);

  if (nic_id == nip->my_nic_id) {
    return;
  }

#if MXOED_DEBUG
  printf("got pkt from nic_id %02x%04x, sn=%d ",
      nic_half, ntohl(pkt->nic_id_lo), serial);
#endif

  index = get_peer_index(nip, nic_id);
  if (index == -1) {
#if MXOED_DEBUG
  printf("new peer\n");
#endif
    add_peer(nip, nic_id, serial, gw);
    nip->bc_count = BROADCAST_COUNT;

    /* make sure interval is at most BROADCAST_INTERVAL */
    if (nip->bc_interval > BROADCAST_INTERVAL) {
      nip->bc_interval = BROADCAST_INTERVAL;
    }
  } else {

    /* new serial number means he likely does not know me */
    if (nip->peer_serial[index] != serial) {
#if MXOED_DEBUG
      printf("known, but serial changed\n");
#endif

      /* record new serial # and broadcast my ID */
      nip->peer_serial[index] = serial;
      nip->bc_count = BROADCAST_COUNT;
      if (nip->bc_interval > BROADCAST_INTERVAL) {
	nip->bc_interval = BROADCAST_INTERVAL;
      }
    } else {
#if MXOED_DEBUG
      printf("already known\n");
#endif
    }
  }
  return;
}

#if MX_OS_MACOSX
#include <sys/param.h>
#define MAX_COMPUTERNAME_LENGTH (MAXHOSTNAMELEN - 1)
#define DWORD size_t
#endif

#if MX_OS_WINNT || MX_OS_MACOSX
void
set_nic_hostname(struct nic_info *nip)
{
  char name[MX_MAX_STR_LEN+1];
  char name2[MAX_COMPUTERNAME_LENGTH+3+1];
  DWORD name2_len;

  if (mx_nic_id_to_hostname(nip->my_nic_id, name) == MX_SUCCESS) {
    if (MAX_COMPUTERNAME_LENGTH > MX_MAX_STR_LEN) {
      name2_len = MX_MAX_STR_LEN;
    } else {
      name2_len = MAX_COMPUTERNAME_LENGTH+3+1;
    }
    name2_len -= 3; /* to allow for :99 mx boards */
    if ((strncmp(name, "localhost", strlen("localhost")) == 0)
#if MX_OS_WINNT
	&& GetComputerName(name2, &name2_len)
#else
	&& (gethostname(name2, name2_len - 1) == 0)
#endif
	) {
      mx_snprintf(name, sizeof(name), "%s:%d", name2, nip->nic_index);
      mx_raw_set_hostname(nip->raw_ep, name);
#if MXOED_DEBUG
      printf("[%s]\n", name);
#endif
    }
  }
}
#endif

void
fill_nic_info(
  struct nic_info *nip)
{
  mx_return_t mxrc;
  uint32_t nic_half;
  uint32_t board_num, type;

  mxrc = mx_board_number_to_nic_id(nip->nic_index, &nip->my_nic_id);
  if (mxrc != MX_SUCCESS) {
    fprintf(stderr, "Error getting nic_id for NIC %d\n", nip->nic_index);
    exit(1);
  }

  memset(&nip->outpkt, 0, sizeof(nip->outpkt));
  nip->outpkt.dest_mac_high32 = 0xFFFFFFFF;
  nip->outpkt.dest_mac_low16 = 0xFFFF;
  nip->outpkt.src_mac_high16 = htons((nip->my_nic_id >> 32) & 0xFFFF);
  nip->outpkt.src_mac_low32 = htonl(nip->my_nic_id & 0xFFFFFFFF);
  nip->outpkt.proto = htons(0x86DF);
  nip->outpkt.pkt_type = 1;		

  /* put my nic_id in outbound packet */
  nic_half = (nip->my_nic_id >> 32) & 0xFFFFFFFF;
  nip->outpkt.nic_id_hi = htonl(nic_half);
  nic_half = nip->my_nic_id & 0xFFFFFFFF;
  nip->outpkt.nic_id_lo = htonl(nic_half);

  /* assign a random serial number for this invocation */
  nip->my_serial = random();
  nip->outpkt.serial = htonl(nip->my_serial);

  board_num = nip->nic_index;
  mx_get_info(NULL, MX_NET_TYPE, &board_num, sizeof(board_num), &nip->net_type, sizeof(nip->net_type));
}

void *
nic_thread(
  void *vnip)
{
  struct nic_info *nip;
  int elapsed;
  int rc;

  nip = vnip;

  fill_nic_info(nip);

#if MX_OS_WINNT || MX_OS_MACOSX
  set_nic_hostname(nip);
#endif

  write_peers_to_nic(nip);	/* write self to NIC peer table */

  nip->bc_count = BROADCAST_COUNT;
  nip->bc_interval = 0;

  while (!nip->die) {

    /* If broadcasts left to do and interval expired, send one now */
    if (nip->bc_count > 0 && nip->bc_interval <= 0 && nip->net_type == MX_NET_ETHER) {
      broadcast_my_id(nip);
      --nip->bc_count;
      if (nip->bc_count > 0) {
	nip->bc_interval = BROADCAST_INTERVAL;
      } else {
	nip->bc_count = 1;
	nip->bc_interval = LONG_BROADCAST_INTERVAL;
      }
    }

    rc = check_for_packet(nip, &elapsed);

    if (nip->bc_interval > 0) {
      nip->bc_interval -= elapsed;
    }

    if (rc > 0) {
      process_pkt(nip);
    }
  }
  return NULL;
}


/*
 * Open NICs
 */
void
open_all_nics()
{
  int i;
  int rc;
  int num_nics;
  struct nic_info *nip;
  struct nic_info *nip0;
  mx_raw_endpoint_t ep;
  MX__THREAD_T tid;

  num_nics = 0;
  for (i=0; i<MAX_NICS; ++i) {

    /* open raw endpoint for this NIC */
    rc = mx_raw_open_endpoint(i, NULL, 0, &ep);
    if (rc != 0) {
      continue;
    }

    /* allocate NIC info struct */
    nip = (struct nic_info *) calloc(sizeof(*nip), 1);
    if (nip == NULL) {
      fprintf(stderr, "Error allocating NIC info struct\n");
      exit(1);
    }
    nip->raw_ep = ep;
    nip->nic_index = i;
    nip->die = 0;

    /* the first NIC will be handled in main thread */
    if (num_nics > 0) {
      rc = MX__THREAD_CREATE(&tid, nic_thread, nip);
      if (rc != 0) {
	fprintf(stderr, "Error creating thread for NIC %d\n", i);
	exit(1);
      }
    } else {
      nip0 = nip;
    }
    ++num_nics;
  }

  /* Error out if no NICs opened */
  if (num_nics == 0) {
    fprintf(stderr, "No NICs found\n");
    exit(1);
  } else {
    nic_thread(nip0);
  }
}

#if MX_OS_WINNT

LPTSTR service_name = "MXoED";
HANDLE service_handle;
SERVICE_STATUS service_status;
HANDLE thread_handle[MAX_NICS];
struct nic_info ni[MAX_NICS];

DWORD WINAPI mapper_thread(LPVOID p)
{
  nic_thread(p);
  return 0;
}

int board_count(void)
{
  char buff[80];
  HANDLE h;
  int i;

  i = 0;
  sprintf(buff, "\\\\.\\mx%d", i);
  h = CreateFile(buff, GENERIC_READ | GENERIC_WRITE, 0, NULL,
		 OPEN_EXISTING, FILE_ATTRIBUTE_SYSTEM, NULL);
  while (h != INVALID_HANDLE_VALUE) {
    CloseHandle(h);
    ++i;
    sprintf(buff, "\\\\.\\mx%d", i);
    h = CreateFile(buff, GENERIC_READ | GENERIC_WRITE, 0, NULL,
		   OPEN_EXISTING, FILE_ATTRIBUTE_SYSTEM, NULL);
  }
  return i;
}

VOID WINAPI service_handler(DWORD fdwControl)
{
  int count;
  int i;

  switch(fdwControl) {
  case SERVICE_CONTROL_STOP:
    service_status.dwCurrentState = SERVICE_STOP_PENDING;
    service_status.dwCheckPoint++;
    service_status.dwWaitHint = 5000;
    SetServiceStatus(service_handle, &service_status);

    /* kill mapper threads */
    count = board_count ();
    for (i = 0; i < count; ++i) {
      ni[i].die = 1;
      broadcast_my_id(&ni[i]);
    }
    if (count > 0) {
      WaitForMultipleObjects (count, thread_handle, TRUE, INFINITE);
    }
    for (i = 0; i < count; ++i) {
      mx_raw_close_endpoint(ni[i].raw_ep);
    }

    service_status.dwCurrentState = SERVICE_STOPPED;
    service_status.dwCheckPoint = 0;
    service_status.dwWaitHint = 0;
    SetServiceStatus(service_handle, &service_status);
    break;
  case SERVICE_CONTROL_PAUSE:
    break;
  case SERVICE_CONTROL_CONTINUE:
    break;
  case SERVICE_CONTROL_INTERROGATE:
    service_status.dwCurrentState = SERVICE_RUNNING;
    service_status.dwCheckPoint = 0;
    service_status.dwWaitHint = 0;
    SetServiceStatus(service_handle, &service_status);
    break;
  case SERVICE_CONTROL_SHUTDOWN:
    break;
  default:
    break;
  }
}

VOID WINAPI service_main(DWORD dwArgc, LPTSTR* lpszArgv)
{
  int count;
  int i;
  mx_raw_endpoint_t ep;

  service_handle = RegisterServiceCtrlHandler(service_name,
					      service_handler);
  service_status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
  service_status.dwCurrentState = SERVICE_START_PENDING;
  service_status.dwControlsAccepted = SERVICE_ACCEPT_STOP;
  service_status.dwWin32ExitCode = NO_ERROR;
  service_status.dwServiceSpecificExitCode = 0;
  service_status.dwCheckPoint = 0;
  service_status.dwWaitHint = 1000;
  SetServiceStatus(service_handle, &service_status);

  /* spawn thread per mapper */
  count = board_count ();
  for (i = 0; i < count; ++i)
  {
    mx_raw_open_endpoint(i, NULL, 0, &ep);

    ni[i].raw_ep = ep;
    ni[i].nic_index = i;
    ni[i].die = 0;

    thread_handle[i] = CreateThread (NULL, 0, mapper_thread, ni+i,
				     0 /*CREATE_SUSPENDED*/, NULL);
  }

  service_status.dwCurrentState = SERVICE_RUNNING;
  service_status.dwCheckPoint++;
  service_status.dwWaitHint = 0;
  SetServiceStatus(service_handle, &service_status);
}

void windows_service(void)
{
  SERVICE_TABLE_ENTRY service_table[] =
  {
    { service_name, service_main },
    { NULL, NULL}
  };

  StartServiceCtrlDispatcher(service_table);
}
#endif

int
main(
  int argc,
  char *argv[])
{
#if MX_OS_WINNT
  if (argc == 2 &&
      strcmp("--service", argv[1]) == 0) {
    mx_init();
    windows_service();
    return 0;
  }
  else {
#endif
  srandom((unsigned int)time(NULL));
#if !MX_OS_WINNT
  setlinebuf(stdout);
  setlinebuf(stderr);
#endif

  /* init mx */
  mx_init();
  open_all_nics();
  exit(0);
#if MX_OS_WINNT
  }
#endif
}
